//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------
texture2D labTexture;
texture2D noiseTexture;
float2 viewport;

//Samplers -------------------------------------------------------------------------

sampler2D labtex = sampler_state
{
	Texture = <labTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
};

sampler2D noisetex = sampler_state
{
	Texture = <noiseTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
};

//----------------------------------------------------------------------------------

// Pixel Shader input structure (= Vertex Shader Output)
struct PixelShaderInput
{
    float2 texC			: TEXCOORD0;
};


////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 PCA1PS(PixelShaderInput input) : COLOR0
{
	// get color of pixel (in L*a*b* space)	
	float4 color = tex2D(labtex, input.texC); 

	// convert from [0,+1] to [0,510] and shift to [-255,255]:
	//color.xyz = (color.xyz*510.0f)-255.0f;

	// convert from [0,+1] to [-0.5,0.5] 
	color.xyz = (color.xyz)-0.5f;

	// get neighbour coordinate indicated by gaussian texture
	float4 neighbourCoord = tex2D(noisetex, input.texC); 
	neighbourCoord.xy = neighbourCoord.xy*viewport;

	// get color of neighbour indicated by gaussian texture coordinate
	float4 neighbourColor = tex2D(labtex, neighbourCoord.xy); 
	//neighbourColor.xyz = (neighbourColor.xyz*510.0f)-255.0f;
	neighbourColor.xyz = (neighbourColor.xyz)-0.5f;

	// calculate relative loss of color contrast (using distance in L*a*b* space)
	//_______________________________________________________
	//             || Pixel_proj - Neighbour_proj ||
	// loss =  1- -----------------------------------
	//             ||    Pixel   -   Neighbour    ||	
	//_______________________________________________________


	float DL = color.x-neighbourColor.x;			//	|| (x1)   (x2) ||   || (x1-x2) ||
	float Da = color.y-neighbourColor.y;			//	|| (y1) - (y2) || = || (y1-y2) ||
	float Db = color.z-neighbourColor.z;			//	|| (z1)   (z2) ||   || (z1-z2) ||

	float DL2 = pow(DL,2.0f);			//	|| (x1-x2) ||
	float Da2 = pow(Da,2.0f);			//	|| (y1-y2) || = sqrt( dx^2 + dy^2 +dz^2)
	float Db2 = pow(Db,2.0f);			//	|| (z1-z2) ||

	float DLab = sqrt(DL2+Da2+Db2); // = distance between pixelColor and neighbourColor in L*a*b* space
	float Dproj = sqrt(DL2+Db2); // = distance between pixelColor and neighbourColor in L*a*b* space projected on b*-plane (a*=y=0)
	
	float factor = 1.0f;
	if (DLab != 0.0f) {					//      || Pixel_proj - Neighbour_proj ||
		factor = abs(Dproj)/DLab;		// =   -----------------------------------
	}									//      ||    Pixel   -   Neighbour    ||

	float loss = 1.0f-factor; // relative loss of color contrast between pixel and neighbour
	
	float a = loss*Da;			// weighted direction of loss = loss*(DL, Da, Db)
	float b = loss*Db;			// but later we will compute the EigenVector of matrix [aa ab; ba bb] 
								//so we store only what we need: a*a, a*b and b*b
	float4 outColor = float4(a*a,a*b,b*b, 1.0f); // Since we preserve the lightness coordinate (i.e., L*) of the original colors to 
													// avoid polarity reversal, it is sufficient to compute the direction that maximizes 
													// contrast loss on the chromaticity plane.
	
	
	//outColor.xyz = outColor.xyz *5.0f;
	
	// shift [-0.5, 0.5] to [0 1]
	outColor.xyz = (outColor.xyz+0.5f); 


	//debug:
	if((outColor.x > 1.0f) || (outColor.y > 1.0f) || (outColor.z > 1.0f))
	{
		outColor.xyz = float3(0.0f, 0.0f, 1.0f);
	}
	else if((outColor.x < 0.0f) || (outColor.y < 0.0f) || (outColor.z < 0.0f))
	{
		outColor.xyz = float3(0.0f, 1.0f, 0.0f);
	}

	outColor.w = 1.0f;
	return outColor;
}


technique PCA1
{
    pass Pass1
    {		
		PixelShader = compile ps_2_0 PCA1PS();
    }
}

