//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------
float4x4 myWorldViewProjectionMatrix : WorldViewProjection;
float4x4 myWorldMatrix : World;
float3 CameraPosition = 0.0f;
float3 DirectLightVector;

texture2D Front;
texture2D Back;
texture2D Transfer;
texture2D TransferAlpha;
texture3D Volume;

texture2D selectedPositions;

float3 StepSize;
float baseSampleDist = .5f;
float sampleDist = 1.0f;
int Iterations;

int Side = 2;
float Slice = 0.5f;

float currentControlPoint;
float prevControlPoint;
float nextControlPoint;

float4 ScaleFactor;
float2 textureSize;

bool frontToBack;
bool shading;
bool useAlpha;
bool useTransferColors;

//Samplers -------------------------------------------------------------------------

sampler2D FrontSampler = sampler_state
{
	Texture = <Front>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				 
    AddressV = CLAMP;				 
};

sampler2D BackSampler = sampler_state
{
	Texture = <Back>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				
    AddressV = CLAMP;				
};

sampler1D TransferSampler = sampler_state
{
	Texture = <Transfer>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU  = CLAMP;
    AddressV  = CLAMP;
};

sampler1D TransferSamplerAlpha = sampler_state
{
	Texture = <TransferAlpha>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU  = CLAMP;
    AddressV  = CLAMP;
};

sampler2D SelectedPosSampler = sampler_state
{
	Texture = <selectedPositions>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				 
    AddressV = CLAMP;	
};

sampler3D VolumeSampler = sampler_state
{
	Texture = <Volume>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				
    AddressV = CLAMP;			
    AddressW = CLAMP;
};

//----------------------------------------------------------------------------------

// Vertex Shader input structure
struct VertexShaderInput
{
    float4 Position : POSITION0;
    float2 texC		: TEXCOORD0;
};

// Pixel Shader input structure (= Vertex Shader Output)
struct VertexShaderOutput
{
    float4 Position		: POSITION0;
    float3 texC			: TEXCOORD0;
    float4 pos			: TEXCOORD1;    
	float3 viewDirection : TEXCOORD2;
};

//lighting values--------------------------------------
float3 Ka = float3(1.0f, 1.0f, 1.0f); //ambient
float3 Kd = float3(1.0f, 1.0f, 1.0f); //diffuse
float3 Ks = float3(1.0f, 1.0f, 1.0f); //specular
float3 DirectLightColor = float3(1.0f, 1.0f, 1.0f);
float3 AmbientLightColor = float3(0.5f, 0.5f, 0.5f);
float3 SpecularLightColor = float3(0.2f, 0.2f, 0.2f);
float shininess = 64.0f; 

float3 CalculateAmbient(float3 materialColor, float3 lightColor)
{
	float3 AmbientLight =  Ka * materialColor * lightColor;
	return AmbientLight;
}

float3 CalculateDiffuse(float3 materialColor, float3 directLightColor, float3 diff)
{
	float3 DiffuseLight =  Kd * materialColor * directLightColor * diff;
	return DiffuseLight;
}

float3 LightingCombine(float3 AmbientLight, float3 DiffuseLight, float3 SpecularLight)
{
	float3 CombinedLight = AmbientLight + DiffuseLight + Ks * SpecularLight;
	return CombinedLight;
}


////////////////////////////////////////////////////////////////////////////////////
//Vertex Shader - Main Function
////////////////////////////////////////////////////////////////////////////////////
VertexShaderOutput PositionVS(VertexShaderInput input)
{
    VertexShaderOutput output;

    output.Position = mul(input.Position * ScaleFactor, myWorldViewProjectionMatrix);
    
    output.texC = input.Position;
    output.pos = output.Position;
    
    // for specular part:
    float4 worldPosition = normalize(mul(input.Position, myWorldMatrix));
	output.viewDirection = CameraPosition - worldPosition;

    return output;
}

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 RayCastSimplePS(VertexShaderOutput input) : COLOR0
{ 
	//calculate projective texture coordinates
	//used to project the front and back position textures onto the cube

	float2 texC = input.pos.xyz/= input.pos.w;
	texC.x =  0.5f*texC.x + 0.5f; 
	texC.y = -0.5f*texC.y + 0.5f;  

    //correct texture coordinates
	float2 s = (textureSize.xy - 1.0f)/textureSize.xy;
	float2 o = 0.5f / textureSize.xy;
	texC.x = (s.x)*texC.x + o.x; 
	texC.y = (s.y)*texC.y + o.y;
	
	//process selected position
	//float3 maskedPos = tex2D(SelectedPosSampler, texC).xyz;
	//return float4(maskedPos, 1);

	//front and back position
	float3 front = tex2D(FrontSampler, texC).xyz;
    float3 back = tex2D(BackSampler, texC).xyz;

    float3 direction = 0;
    float4 pos = 0;

    if(frontToBack)
    {
		direction = normalize(back - front);
		pos = float4(front, 0);
    }
    else
    {
		direction = normalize(front - back);
		pos = float4(back, 0);
	}
        
    float4 oldValue = 0;
    float4 newValue = 0;
    float4 finalValue = 0;

	float4 finalWeight = 0.0f;
    float3 sampleWeight = 0.0f;
	float sampleAlpha = 1.0f;
	
    float value = 0.0f;
	
	float3 Step = direction * StepSize;
	
	float adaptiveStep = 1.0f;

    for(int i = 0; i < Iterations; i++)
    {
		pos.w = 0;

		float sampleTheta = 0.0f;
	    float3 sampleBeta = 0.0f;
		float sampleCurrAlpha = 0.0f;
		
		//get density value
		value = tex3Dlod(VolumeSampler, pos).a;

		//theta
		float valueScaled = value * 255.0f;
		//calculate color weight Theta(m,i) for the current control point and surrent sample
		if((prevControlPoint) == 0.0f && (valueScaled < currentControlPoint)) //first cp
		{
		  sampleTheta = 0.0f; //fist != curr
		  
		}else if((nextControlPoint == 255.0f) && (valueScaled > currentControlPoint)) //last cp
		{
		   sampleTheta = 0.0f; //curr != last
		}
		else
		{
	      if(valueScaled < currentControlPoint && valueScaled > prevControlPoint)
	       {
		      sampleTheta = (valueScaled - prevControlPoint)/(currentControlPoint - prevControlPoint);
			   
	       }
	      else if(valueScaled >= currentControlPoint && valueScaled < nextControlPoint)
	       {
	          sampleTheta = 1 - (valueScaled - currentControlPoint)/(nextControlPoint - currentControlPoint);
			  
           }
		  else 
		   { 
		      sampleTheta = 0.0f; 
		   }

	   
	   }
	   //end theta

	   //shading
	   if(shading)
		{
			//get gradient
			float3 gradient = tex3Dlod(VolumeSampler, pos).rgb;
			gradient.xyz = normalize(gradient.xyz*2.0f - 1.0f);
			//gradient.xyz = (gradient.xyz*2.0f - 1.0f);
			//float magnitude = length(gradient);
			//gradient = float3(magnitude, magnitude, magnitude);

			//float3 LightDirection = normalize(DirectLightVector);
		    //float3 ViewDirection = normalize(input.viewDirection);
			float3 LightDirection = normalize(DirectLightVector - pos.xyz); 
			float3 ViewDirection = normalize(CameraPosition - pos.xyz);
			float3 Half = normalize(LightDirection + ViewDirection);
			float diff = saturate(dot(LightDirection, gradient));
					
				
			if(useTransferColors)
			{
				//get color from transferTexture
				newValue = tex1Dlod(TransferSampler, value);
			
			}
			else
			{
				// density values produce gray picture
				newValue = float4(value, value, value, value);
			}

			//calculate color for ambient lighting
			//float3 AmbientLight = CalculateAmbient(float4(1,1,1,1), AmbientLightColor);
			float3 AmbientLight = CalculateAmbient(newValue, AmbientLightColor);
			
			//calculate diffuse lighting
			//float3 DiffuseLight = CalculateDiffuse(float4(1,1,1,1), DirectLightColor, diff);
			float3 DiffuseLight = CalculateDiffuse(newValue, DirectLightColor, diff);
			
			//calculate specular Lighting 
			//phong
			//float3 Reflect = normalize(2 * diff * gradient - LightDirection);
			//float3 SpecularLight = SpecularLightColor * pow(saturate(dot(Reflect, ViewDirection)), 64.0f);
				
		    //blinn
			float specularTerm = diff * pow(saturate(dot(Half, gradient)), shininess);
			float3 SpecularLight = SpecularLightColor * specularTerm;

			//combine the light-values
			newValue.rgb = LightingCombine(AmbientLight, DiffuseLight, SpecularLight);
			 // newValue.rgb = LightingCombine(AmbientLight, float3(0.0f, 0.0f, 0.0f), float3(0.0f, 0.0f, 0.0f));

			sampleBeta = newValue.rgb;
					
		}//end shading
		else
		{	
		    sampleBeta = 1.0f;
			if(useTransferColors)
			{
				// get color from transferTexture
				newValue = tex1Dlod(TransferSampler, value);
			}
			else
			{
				// density values produce gray picture
				newValue = float4(value, value, value, value);
				
			}
			
		}
		
		if(useAlpha)
		{
			//alpha from transferfunction
			newValue.a = tex1Dlod(TransferSampler, value).a;
		}
		else
		{
			// set alpha from density values
			newValue.a = value;
		}

		newValue.a = 1 - pow((1 - newValue.a), sampleDist / baseSampleDist); 
				
		if(frontToBack)
		{					
			//front to back blending
			//finalValue.rgb = newValue.rgb * newValue.a *(1 - oldValue.a) + oldValue.rgb;
			finalValue.a = newValue.a *(1 - oldValue.a)+ oldValue.a; 
	
		}
		else
		{
			//back to front blending ... ATTENTION: you may not stop at alpha > 0.99 for back to front!
			//finalValue.rgb = newValue.rgb * newValue.a + oldValue.rgb * (1 - newValue.a);			
			finalValue.a = newValue.a + oldValue.a *(1 - newValue.a); 

	     }
		
		oldValue = finalValue;


		//set weight alpha 
		//sampleCurrAlpha = newValue.a;
	    //finalValue.a = newValue.a + oldValue.a *(1 - newValue.a); 
		//oldValue = finalValue;

		if(value > 0.0f)
		{
			adaptiveStep = 1.0f;
			//adaptiveStep = newValue.a*5.0f;
			//adaptiveStep = 1.0f/(value*10.0f);
		}
		else
		{
			adaptiveStep = 1.0f;
		}
		
		//break from the loop when alpha gets high enough
		if((finalValue.a > 0.99f)&& (frontToBack == true))
		{
			i = Iterations;	
		}
		else
			{
			
			//advance the current position
			pos.xyz += Step * adaptiveStep;
			
			//break if the position is greater than <1, 1, 1>
			if(pos.x > 1.0f || pos.y > 1.0f || pos.z > 1.0f)
			{
				i = Iterations;
			}
			
		}
		sampleWeight = sampleTheta * sampleBeta * finalValue.a;
			
		finalWeight += float4(sampleWeight.x,sampleWeight.y,sampleWeight.z, 1.0f);
		
		//back to front
		//sampleAlpha *= (1 - sampleCurrAlpha); 

    }//end for iterations
	    
   
	//shift final weight in [0 ; 1]
	//finalWeight.xyz = (finalWeight.xyz + 1.0f)/2.0f;
    return finalWeight;
	
	
}


//----------------------------------------------------------------------------------
//Render Scene
//----------------------------------------------------------------------------------
technique CalculateColorWeights
{
    pass Pass1
    {		
        VertexShader = compile vs_3_0 PositionVS();
        PixelShader = compile ps_3_0 RayCastSimplePS();
    }
}

