//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------

float3 DirectLightVector; 

//set during draw------------------------------
float4x4 myWorldViewProjectionMatrix : WorldViewProjection;
float4x4 myWorldMatrix : World;
float3 CameraPosition;

texture2D Front;
texture2D Back;
texture2D TransferAlpha;
texture2D CVDTransfer;
texture3D Volume;
texture3D Gradients;

float3 StepSize;
int Iterations;
float4 ScaleFactor;
float2 textureSize; 

bool frontToBack;
bool shading;
bool useAlpha;
bool useCVDTransferColors;

// RGB to L*a*b*
float4x4 rotmat; 
float4x4 mRGBtoXYZ; // RGB to XYZ matrix conversion

// L*a*b* to RGB
float4x4 rotmatBack;
float4x4 mXYZtoRGB; // XYZ to RGB matrix conversion

//set during draw end----------------------------

float baseSampleDist = .5f;
float sampleDist = 1.0f;




//Samplers -------------------------------------------------------------------------

sampler2D FrontSampler = sampler_state
{
	Texture = <Front>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				
    AddressV = CLAMP;				
};

sampler2D BackSampler = sampler_state
{
	Texture = <Back>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				
    AddressV = CLAMP;				
};

sampler1D CVDTransferSampler = sampler_state
{
	Texture = <CVDTransfer>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU  = CLAMP;
    AddressV  = CLAMP;
};

sampler1D TransferSamplerAlpha = sampler_state
{
	Texture = <TransferAlpha>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU  = CLAMP;
    AddressV  = CLAMP;
};

sampler3D VolumeSampler = sampler_state
{
	Texture = <Volume>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				
    AddressV = CLAMP;				
    AddressW = CLAMP;
};


//----------------------------------------------------------------------------------

// Vertex Shader input structure
struct VertexShaderInput
{
    float4 Position : POSITION0;
    float2 texC		: TEXCOORD0;
};

// Pixel Shader input structure (= Vertex Shader Output)
struct VertexShaderOutput
{
    float4 Position		: POSITION0;
    float3 texC			: TEXCOORD0;
    float4 pos			: TEXCOORD1;    
	float3 viewDirection : TEXCOORD2;
};

//lighting values--------------------------------------
float3 Ka = float3(1.0f, 1.0f, 1.0f); //ambient
float3 Kd = float3(1.0f, 1.0f, 1.0f); //diffuse
float3 Ks = float3(1.0f, 1.0f, 1.0f); //specular
float3 DirectLightColor = float3(1.0f, 1.0f, 1.0f);
float3 AmbientLightColor = float3(0.5f, 0.5f, 0.5f);
float3 SpecularLightColor = float3(0.2f, 0.2f, 0.2f);
float shininess = 64.0f; 

float3 CalculateAmbient(float3 materialColor, float3 lightColor)
{
	float3 AmbientLight =  Ka * materialColor * lightColor;
	return AmbientLight;
}

float3 CalculateDiffuse(float3 materialColor, float3 directLightColor, float3 diff)
{
	float3 DiffuseLight =  Kd * materialColor * directLightColor * diff;
	return DiffuseLight;
}

float3 LightingCombine(float3 AmbientLight, float3 DiffuseLight, float3 SpecularLight)
{
	float3 CombinedLight = AmbientLight + DiffuseLight + Ks * SpecularLight;
	return CombinedLight;
}

// color space transformations------------------------------------------------------

float F(float p)
{
	if (p>0.008856f)
	{
		return pow(p,1.0f/3.0f);
	}
	else
	{
		return (p*903.3f + 16.0f)/116.0f;
	}	
}

float invF(float p)
{
	float p3 = pow(p,3.0f);
	if (p3 > 0.008856f)
	{
		return p3;
	}
	else
	{
		return (116.0f * p - 16.0f)/903.3f;
	}
}

// Convert input CIE XYZ  to  L*a*b* color space
float3 XYZToLab(float3 vec3)
{
	// divide input (X, Y, Z) by reference white
	float3 white = ( 0.9642f, 1.0f, 0.8249f); // Reference white Adobe Photoshop

	float xr = vec3.x/white.x;
	float yr = vec3.y/white.y;
	float zr = vec3.z/white.z;

	float fx = F(xr);
	float fy = F(yr);
	float fz = F(zr);

	float L = 116.0f * fy - 16.0f;
	float a = 500.0f * (fx - fy);
	float b = 200.0f * (fy - fz);

	return float3(L, a, b);
}

// Convert input L*a*b* to CIE XYZ color space
float3 LabToXYZ(float3 vec3)
{
	float L = vec3.x;
	float a = vec3.y;
	float b = vec3.z; 

	float fy = (L+16.0f)/116.0f;
	float fx = (a/500.0f) + fy;
	float fz = fy - (b/200.0f);

	float xr = invF(fx); 
	float yr = invF(fy);
	float zr = invF(fz);

	// multiply by reference white
	float3 white = ( 0.9642f, 1.0f, 0.8249f); // Reference white Adobe Photoshop
	
	float X = xr * white.x;
	float Y = yr * white.y;
	float Z = zr * white.z;	

	return float3(X, Y, Z);
}

float gammaCorrection(float value)
{
	float ret;
	if (value <= 0.04045f) {
		ret = value/12.92f;
	} else {
		ret = pow(((value + 0.055f)/1.055f),2.4f);
	}
	return ret;
}

float InvGammaCorrection(float value)
{
	float ret;
	if (value <= 0.0031308f) {
		ret = value*12.92f;
	} else {
		ret = 1.055f * pow(value,1.0f/2.4f) - 0.055f;
	}
	return ret;
}

float3 RGBGammaCorrection(float3 rgbcol)
{
	float X = gammaCorrection(rgbcol.x);
	float Y = gammaCorrection(rgbcol.y);
	float Z = gammaCorrection(rgbcol.z);
	float3 result = float3(X,Y,Z);
	return result;
}

float3 InvRGBGammaCorrection(float3 rgbcol)
{
	float X = InvGammaCorrection(rgbcol.x);
	float Y = InvGammaCorrection(rgbcol.y);
	float Z = InvGammaCorrection(rgbcol.z);
	return float3(X,Y,Z);
}

// cvd friendly color blending------------------------------------------------------

float4 blend(float4 newColor, float4 oldColor)
{
	float4 blended = float4(0.0f, 0.0f, 0.0f, 0.0f);
	
	if(frontToBack)
	{
		// blend L* (=x) for Luminance Consistency
		//blended.x = newColor.x * newColor.a *(1 - oldColor.a) + oldColor.x;

		// blend b* (=z)
		//blended.z = newColor.z * newColor.a *(1 - oldColor.a) + oldColor.z;

		//blended.y = 0; // project to plane

		
		blended.xyz = newColor.xyz * newColor.a *(1 - oldColor.a) + oldColor.xyz;

		blended.a = newColor.a *(1 - oldColor.a) + oldColor.a; 
	}
	else
	{
		// blend L* (=x) for Luminance Consistency
		//blended.x = newColor.x * newColor.a + oldColor.x * (1 - newColor.a);	
		
		// blend b* (=z)
		//blended.z = newColor.z * newColor.a + oldColor.z * (1 - newColor.a);	

		//blended.y = 0; // project to plane

		blended.xyz = newColor.xyz * newColor.a + oldColor.xyz * (1 - newColor.a);
				
		blended.a = newColor.a + oldColor.a *(1 - newColor.a); 
	}

	return blended;
}

float4 cvdFriendlyBlended(float4 newColor, float4 oldColor)
{
	// do gamma correction
	newColor.xyz = RGBGammaCorrection(newColor.xyz);
	oldColor.xyz = RGBGammaCorrection(oldColor.xyz);

	// convert from RGB to XYZ
	newColor.xyz = mul(float4(newColor.xyz, 1.0f), mRGBtoXYZ); //mit alpha oder ohne?
	oldColor.xyz = mul(float4(oldColor.xyz, 1.0f), mRGBtoXYZ);
				
	// convert from XYZ to Lab
	newColor.xyz = XYZToLab(newColor.xyz);	
	oldColor.xyz = XYZToLab(oldColor.xyz);	
		
	// rotate according to specified color vision deficiency
	newColor.xyz = mul(float4(newColor.xyz, 1.0f), rotmat).xyz;
	oldColor.xyz = mul(float4(oldColor.xyz, 1.0f), rotmat).xyz;
	
	// blend colors on plane representing reduced color space
	float4 blendedColor = blend(newColor, oldColor); 

	// rotate according to specified color vision deficiency, rotate plane back
	blendedColor.xyz = mul(float4(blendedColor.xyz, 1.0f), rotmatBack).xyz;
		
	// convert from L*a*b* back to XYZ
	blendedColor.xyz = LabToXYZ(blendedColor.xyz);
				
	// convert from XYZ back to RGB
	blendedColor.xyz = mul(float4(blendedColor.xyz, 1.0f), mXYZtoRGB);
	
	// do inverse gamma correction
	blendedColor.xyz = InvRGBGammaCorrection(blendedColor.xyz);

	return blendedColor;
}

////////////////////////////////////////////////////////////////////////////////////
//Vertex Shader - Main Function
////////////////////////////////////////////////////////////////////////////////////
VertexShaderOutput PositionVS(VertexShaderInput input)
{
    VertexShaderOutput output;

    output.Position = mul(input.Position * ScaleFactor, myWorldViewProjectionMatrix);
    
    output.texC = input.Position;
    output.pos = output.Position;
    
    // for specular part:
    float4 worldPosition = normalize(mul(input.Position, myWorldMatrix));
	output.viewDirection = CameraPosition - worldPosition;

    return output;
}

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 PositionPS(VertexShaderOutput input) : COLOR0
{		
    return float4(input.texC, 1.0f);
}

float4 RayCastSimplePS(VertexShaderOutput input) : COLOR0
{ 
	//calculate projective texture coordinates
	//used to project the front and back position textures onto the cube
	float2 texC = input.pos.xyz/= input.pos.w;

	texC.x =  0.5f*texC.x + 0.5f; 
	texC.y = -0.5f*texC.y + 0.5f;  

	//correct texture coordinates
	float2 s = (textureSize.xy - 1.0f)/textureSize.xy;
	float2 o = 0.5f / textureSize.xy;
	texC.x = (s.x)*texC.x + o.x; 
	texC.y = (s.y)*texC.y + o.y;
	
	
    float3 front = tex2D(FrontSampler, texC).xyz;
    float3 back = tex2D(BackSampler, texC).xyz;
    
    float3 direction = 0;
    float4 pos = 0;
    if(frontToBack)
    {
		direction = normalize(back - front);
		pos = float4(front, 0);
    }
    else
    {
		direction = normalize(front - back);
		pos = float4(back, 0);
	}
 
    //compositing step      
    float4 oldValue = 0;
    float4 newValue = 0;
    float4 finalValue = 0;
    
    float value = 0.0f;
	
	float3 Step = direction * StepSize;
	
	float adaptiveStep = 1.0f;
    
    for(int i = 0; i < Iterations; i++)
    {
		pos.w = 0;
		// get density value
		value = tex3Dlod(VolumeSampler, pos).a;
		
		if(shading)
		{
		    //get gradient
			float3 gradient = tex3Dlod(VolumeSampler, pos).rgb;
			gradient.xyz = normalize(gradient.xyz*2.0f - 1.0f);

			//float3 LightDirection = normalize(DirectLightVector);
		    float3 ViewDirection = normalize(input.viewDirection);
			float3 LightDirection = normalize(DirectLightVector - pos.xyz); 
			//float3 ViewDirection = normalize(CameraPosition - pos.xyz);
			float3 Half = normalize(LightDirection + ViewDirection);
			float diff = saturate(dot(LightDirection, gradient));
			
			//get color from cvd transferTexture
			newValue = tex1Dlod(CVDTransferSampler, (value));
			
			//calculate color for ambient lighting
			float3 AmbientLight = CalculateAmbient(newValue, AmbientLightColor);
			
			//calculate diffuse lighting
			float3 DiffuseLight = CalculateDiffuse(newValue, DirectLightColor, diff);
			
			//calculate specular Lighting 				
		    //blinn phong
			float specularTerm = diff * pow(saturate(dot(Half, gradient)), shininess);
			float3 SpecularLight = SpecularLightColor * specularTerm;

			//combine the light-values
			newValue.rgb = LightingCombine(AmbientLight, DiffuseLight, SpecularLight);				
		}
		else
		{			
			//get color from cvd transferTexture
			newValue = tex1Dlod(CVDTransferSampler, (value));
		}		



		if(useAlpha)
		{
			// alpha from transferfunction
			newValue.a = tex1Dlod(TransferSamplerAlpha, value).a;

			//make it more transparent	
			//newValue.a *= 0.5f;
			
			//Oppacity correction: As sampling distance decreases we get more samples.
			//Therefore the alpha values set for a sampling distance of .5f will be too
			//high for a sampling distance of .25f (or conversely, too low for a sampling
			//distance of 1.0f). So we have to adjust the alpha accordingly.
			//newValue.a = 1 - pow((1 - newValue.a), StepSize / 0.5f);
		}
		else
		{
			// set alpha from density values
			newValue.a = value;
			
			//make it more transparent	TODO: alpha from transferfunction	
			//newValue.a *= 0.1f; //src.a *= 0.02f; //default = 1.0f;
						  //reduce the alpha to have a more transparent result
						  //TODO: this needs to be adjusted based on the step size
						  //i.e. the more steps we take, the faster the alpha will grow	
		}

		//Oppacity correction: As sampling distance decreases we get more samples.
		//Therefore the alpha values set for a sampling distance of .5f will be too
		//high for a sampling distance of .25f (or conversely, too low for a sampling
		//distance of 1.0f). So we have to adjust the alpha accordingly.
		newValue.a = 1 - pow((1 - newValue.a), sampleDist / baseSampleDist); 


		//color composition 	
		
		//with cvd friendly color belnding and luminance constistency:	
		finalValue = cvdFriendlyBlended(newValue,oldValue);	
		
		// without: 		
		/*if(frontToBack)
		{					
			//front to back blending
			finalValue.rgb = newValue.rgb * newValue.a *(1 - oldValue.a) + oldValue.rgb;
			finalValue.a = newValue.a *(1 - oldValue.a) + oldValue.a; 

		}
		else
		{
			//back to front blending ... ATTENTION: you may not stop at alpha > 0.99 for back to front!
			finalValue.rgb = newValue.rgb * newValue.a + oldValue.rgb * (1 - newValue.a);			
			finalValue.a = newValue.a + oldValue.a *(1 - newValue.a); 

	     }	*/
			
		oldValue = finalValue;

		//set iteration
		if(value > 0.0f)
		{
			adaptiveStep = 1.0f;
			//adaptiveStep = newValue.a*5.0f;
			//adaptiveStep = 1.0f/(value*10.0f);
		}
		else
		{
			adaptiveStep = 1.0f;
		}
		
		//break from the loop when alpha gets high enough
		if((finalValue.a >= 0.95f) && (frontToBack == true))
		{
			i = Iterations;	
		}
		else
			{
			
			//advance the current position
			pos.xyz += Step * adaptiveStep;
			
			//break if the position is greater than <1, 1, 1>
			if(pos.x > 1.0f || pos.y > 1.0f || pos.z > 1.0f)
			{
				i = Iterations;
			}
			
		}
    }    
    return finalValue;
}

float4 MaxIntensityProjPS(VertexShaderOutput input) : COLOR0
{ 
	//calculate projective texture coordinates
	//used to project the front and back position textures onto the cube
	float2 texC = input.pos.xyz /= input.pos.w;
	texC.x =  0.5f*texC.x + 0.5f; 
	texC.y = -0.5f*texC.y + 0.5f;  
	
    float3 front = tex2D(FrontSampler, texC).xyz;
    float3 back = tex2D(BackSampler, texC).xyz;    
   
	float3 direction = (back - front);
	direction = normalize(direction);
	float4 pos = float4(front, 0);
            
    float4 finalValue = 0;    
    float4 newValue = 0;
    
    float value = 0.0f;
	
	float3 Step = direction * StepSize;
	
	float3 gradient = 0.0f;
    
    for(int i = 0; i < Iterations; i++)
    {
		pos.w = 0;
		
		// get density value 
		value = tex3Dlod(VolumeSampler, pos).a;
		
		if(useAlpha)
		{
			// alpha from transferfunction
			newValue.a = tex1Dlod(TransferSamplerAlpha, value).a;
			//make it more transparent	
			newValue.a *= 0.5f;			
		}
		else
		{
			// set alpha from density values
			newValue.a = value;
			
		}
		
		if(newValue.a > finalValue.a)
		{	
			if(useCVDTransferColors)
			{
				// get color from transferTexture
				finalValue = tex1Dlod(CVDTransferSampler, (value));
				//finalValue = tex1Dlod(CVDTransferSampler, (newValue.a));
			}
			else
			{
				// density values produce gray picture
				finalValue = float4(newValue.a, newValue.a, newValue.a, newValue.a);
			}
			
			finalValue.rgb *= newValue.a;
			
			// set alpha
			finalValue.a = newValue.a;
			
			gradient = tex3Dlod(VolumeSampler, pos).rgb;
		
		}
		
		//advance the current position
		pos.xyz += Step;
			
		//break if the position is greater than <1, 1, 1>
		if(pos.x > 1.0f || pos.y > 1.0f || pos.z > 1.0f)
		{
			i = Iterations;
		}
    }  
    
    if(shading)
		{
			
			float3 LightDirection = normalize(DirectLightVector); 
			float3 ViewDirection = normalize(input.viewDirection);
			float3 diff = saturate(dot(gradient, LightDirection));			
				
			if(useCVDTransferColors)
			{
				// get color from transferTexture
				newValue = tex1Dlod(CVDTransferSampler, (value));
				//newValue = tex1Dlod(CVDTransferSampler, (finalValue.a));
			}
			else
			{
				// density values produce gray picture
				newValue = float4(finalValue.a, finalValue.a, finalValue.a, finalValue.a);
			}
		
			//calculate color for ambient lighting
			float3 AmbientLight = CalculateAmbient(newValue, AmbientLightColor);
			AmbientLight *= finalValue.a;
			
			//calculate diffuse lighting
			float3 DiffuseLight = CalculateDiffuse(newValue, DirectLightColor, diff);
			//DiffuseLight *= finalValue.a;
			
			//calculate specular Lighting 
			float3 Reflect = normalize(2 * diff * gradient - LightDirection);
			float SpecularLight = pow(saturate(dot(Reflect, ViewDirection)), 64.0f);
			SpecularLight *= finalValue.a;
			
			//combine the light-values
			newValue.rgb = LightingCombine(AmbientLight, DiffuseLight, SpecularLight);
			finalValue = newValue;
		}   
      
    return finalValue;
}

//----------------------------------------------------------------------------------
//Render Scene
//----------------------------------------------------------------------------------
technique RayCastSimple
{
    pass Pass1
    {		
        VertexShader = compile vs_3_0 PositionVS();
        PixelShader = compile ps_3_0 RayCastSimplePS();
    }
}

technique MaximumIntensity
{
    pass Pass1
    {		
        VertexShader = compile vs_3_0 PositionVS();
        PixelShader = compile ps_3_0 MaxIntensityProjPS();
    }
}
