//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------
float4x4 myWorldViewProjectionMatrix : WorldViewProjection;
float4x4 myWorldMatrix : World;
float3 CameraPosition;
float3 DirectLightVector; 

texture2D Front;
texture2D Back;
texture2D Transfer;
texture2D TransferAlpha;
texture2D CVDTransfer;
texture3D Volume;
texture3D Gradients;

float3 StepSize;
float baseSampleDist = .5f;
float sampleDist = 1.0f;
int Iterations;

int Side = 2;
float Slice = 0.5f;

float4 ScaleFactor;

float2 textureSize; 

bool frontToBack;
bool shading;
bool useAlpha;
bool useTransferColors;
bool useCVDTransferColors;

//Samplers -------------------------------------------------------------------------

sampler2D FrontSampler = sampler_state
{
	Texture = <Front>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				
    AddressV = CLAMP;				
};

sampler2D BackSampler = sampler_state
{
	Texture = <Back>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				
    AddressV = CLAMP;				
};

sampler1D TransferSampler = sampler_state
{
	Texture = <Transfer>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU  = CLAMP;
    AddressV  = CLAMP;
};

sampler1D CVDTransferSampler = sampler_state
{
	Texture = <CVDTransfer>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU  = CLAMP;
    AddressV  = CLAMP;
};

sampler1D TransferSamplerAlpha = sampler_state
{
	Texture = <TransferAlpha>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU  = CLAMP;
    AddressV  = CLAMP;
};

sampler3D VolumeSampler = sampler_state
{
	Texture = <Volume>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				
    AddressV = CLAMP;				
    AddressW = CLAMP;
};

sampler3D GradientSampler = sampler_state
{
	Texture = <Gradients>;
	//MinFilter = LINEAR;
	//MagFilter = LINEAR;
	//MipFilter = LINEAR;
	
	AddressU = CLAMP;				
    AddressV = CLAMP;				
    AddressW = CLAMP;
};

//----------------------------------------------------------------------------------

// Vertex Shader input structure
struct VertexShaderInput
{
    float4 Position : POSITION0;
    float2 texC		: TEXCOORD0;
};

// Pixel Shader input structure (= Vertex Shader Output)
struct VertexShaderOutput
{
    float4 Position		: POSITION0;
    float3 texC			: TEXCOORD0;
    float4 pos			: TEXCOORD1;    
	float3 viewDirection : TEXCOORD2;
};

//lighting values--------------------------------------
float3 Ka = float3(1.0f, 1.0f, 1.0f); //ambient
float3 Kd = float3(1.0f, 1.0f, 1.0f); //diffuse
float3 Ks = float3(1.0f, 1.0f, 1.0f); //specular
float3 DirectLightColor = float3(1.0f, 1.0f, 1.0f);
float3 AmbientLightColor = float3(0.5f, 0.5f, 0.5f);
float3 SpecularLightColor = float3(0.2f, 0.2f, 0.2f);
float shininess = 64.0f; 

float3 CalculateAmbient(float3 materialColor, float3 lightColor)
{
	float3 AmbientLight =  Ka * materialColor * lightColor;
	return AmbientLight;
}

float3 CalculateDiffuse(float3 materialColor, float3 directLightColor, float3 diff)
{
	float3 DiffuseLight =  Kd * materialColor * directLightColor * diff;
	return DiffuseLight;
}

float3 LightingCombine(float3 AmbientLight, float3 DiffuseLight, float3 SpecularLight)
{
	float3 CombinedLight = AmbientLight + DiffuseLight + Ks * SpecularLight;
	return CombinedLight;
}

////////////////////////////////////////////////////////////////////////////////////
//Vertex Shader - Main Function
////////////////////////////////////////////////////////////////////////////////////
VertexShaderOutput PositionVS(VertexShaderInput input)
{
    VertexShaderOutput output;

    output.Position = mul(input.Position * ScaleFactor, myWorldViewProjectionMatrix);
    
    output.texC = input.Position;
    output.pos = output.Position;
    
    // for specular part:
    float4 worldPosition = normalize(mul(input.Position, myWorldMatrix));
	output.viewDirection = CameraPosition - worldPosition;

    return output;
}

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 PositionPS(VertexShaderOutput input) : COLOR0
{		
    return float4(input.texC, 1.0f);
}

float4 RayCast(VertexShaderOutput input)
{ 
	//calculate projective texture coordinates
	//used to project the front and back position textures onto the cube
	float2 texC = input.pos.xy /= input.pos.w;
	texC.x =  0.5f*texC.x + 0.5f; 
	texC.y = -0.5f*texC.y + 0.5f;  

    float3 front = tex2D(FrontSampler, texC).xyz;
    float3 back = tex2D(BackSampler, texC).xyz;
    
    float3 dir = normalize(back - front);
    float4 pos = float4(front, 0);
    
    float4 dst = float4(0, 0, 0, 0);
    float4 src = 0;
    
    float value = 0;
	
	float3 Step = dir * StepSize;
    
    for(int i = 0; i < Iterations; i++)
    {
		pos.w = 0;
		value = tex3Dlod(VolumeSampler, pos).r;
				
		src = tex1Dlod(TransferSampler, value); //transferfunction
		
		src.a = value; //transferfunction
		
		//src = (float4)value; // needed without transferfunction
		src.a *= 0.1f; //default = 1.0f;
					  //reduce the alpha to have a more transparent result
					  //this needs to be adjusted based on the step size
					  //i.e. the more steps we take, the faster the alpha will grow	
						
		//Front to back blending
		// dst.rgb = dst.rgb + (1 - dst.a) * src.a * src.rgb
		// dst.a   = dst.a   + (1 - dst.a) * src.a		
		src.rgb *= src.a;
		dst = (1.0f - dst.a)*src + dst;		
		
		//break from the loop when alpha gets high enough
		if(dst.a >= 0.95f)
		{
			i = Iterations;	
		}
		else
			{
			
			//advance the current position
			pos.xyz += Step;
			
			//break if the position is greater than <1, 1, 1>
			if(pos.x > 1.0f || pos.y > 1.0f || pos.z > 1.0f)
			{
				i = Iterations;
			}
		}
    }
    
    return dst*1.0f;
}

float4 SlicePS(VertexShaderOutput input) : COLOR0
{	
	float4 slicePos;
	float4 returnValue;
	if(Side == 0) // step in view-direction
	{
		//slicePos = float4(Slice, Slice, Slice, 0.0f);
		//slicePos = float4(input.texC.x, input.texC.y, input.texC.z, 0.0f);
		returnValue = RayCast(input);
	}
	else if(Side == 1) // step in y-direction
	{
		slicePos = float4(input.texC.x, Slice, input.texC.z, 0.0f);
	}
	else if(Side == 2) // step in x-direction
	{
		slicePos = float4(Slice, input.texC.y, input.texC.z, 0.0f);
	}	
	else if(Side == 3) // step in z-direction
	{
		slicePos = float4(input.texC.x, input.texC.y, Slice, 0.0f);
	}	
	
	if(Side != 0)
	{
		float sliceValue = 0;   
		sliceValue = tex3Dlod(VolumeSampler, slicePos).a;
		//sliceValue *= 10; // make it brighter
		return float4(sliceValue, sliceValue, sliceValue, 1.0f);
	}
	else
	{
		return returnValue;
		//return float4(0.0f, 0.0f, 0.0f, 0.0f);	
	}
		
}

float4 WireFramePS(VertexShaderOutput input) : COLOR0
{
    return float4(1.0f, 0.5f, 0.0f, 0.85f);
}

//draws the front or back positions, or the ray direction through the volume
float4 DirectionPS(VertexShaderOutput input) : COLOR0
{
	float2 texC = input.pos.xy /= input.pos.w;
	texC.x =  0.5f*texC.x + 0.5f; 
	texC.y = -0.5f*texC.y + 0.5f;
	
    float3 front = tex2D(FrontSampler, texC).rgb;
    float3 back = tex2D(BackSampler, texC).rgb;
	
	float4 returnValue;
	
	if(Side == 0)
	{
		returnValue = float4(front, 0.9f);
	}
	else if(Side == 1)
	{
		returnValue = float4(back, 0.9f);
	}
	else
	{
		returnValue = float4(back - front, 0.9f);
	}
    
	return returnValue;
}

float3 filterGradient(float4 pos)
{
//tex3Dlod(VolumeSampler, pos).rgb;
	float sampleSize = 1.0f;
	
	float4 pos1 = pos;
		pos1.x += sampleSize;
	float4 pos2 = pos;
		pos2.x -= sampleSize;
	float4 pos3 = pos;
		pos3.y += sampleSize;
	float4 pos4 = pos;
		pos4.y -= sampleSize;
	float4 pos5 = pos;
		pos5.z += sampleSize;
	float4 pos6 = pos;
		pos6.z -= sampleSize;
	
	float3 filtered = float3(
					tex3Dlod(VolumeSampler, pos1).rgb +
					tex3Dlod(VolumeSampler, pos2).rgb +
					tex3Dlod(VolumeSampler, pos3).rgb +
					tex3Dlod(VolumeSampler, pos4).rgb +
					tex3Dlod(VolumeSampler, pos5).rgb +
					tex3Dlod(VolumeSampler, pos6).rgb +
					tex3Dlod(VolumeSampler, pos).rgb );
					
	filtered /= 7.0f;
	
	return filtered;
	//return tex3Dlod(VolumeSampler, pos).rgb;

		//getSampleAt(x + sampleSize, y, z) - getSampleAt(x - sampleSize, y, z),
		//getSampleAt(x, y + sampleSize, z) - getSampleAt(x, y - sampleSize, z),
		//getSampleAt(x, y, z + sampleSize) - getSampleAt(x, y, z - sampleSize));

}

float4 RayCastSimplePS(VertexShaderOutput input) : COLOR0
{ 
	//calculate projective texture coordinates
	//used to project the front and back position textures onto the cube
	float2 texC = input.pos.xyz/= input.pos.w;

	texC.x =  0.5f*texC.x + 0.5f; 
	texC.y = -0.5f*texC.y + 0.5f;  

	//correct texture coordinates
	float2 s = (textureSize.xy - 1.0f)/textureSize.xy;
	float2 o = 0.5f / textureSize.xy;
	texC.x = (s.x)*texC.x + o.x; 
	texC.y = (s.y)*texC.y + o.y;
	
	
    float3 front = tex2D(FrontSampler, texC).xyz;
    float3 back = tex2D(BackSampler, texC).xyz;
    
    float3 direction = 0;
    float4 pos = 0;
    if(frontToBack)
    {
		direction = normalize(back - front);
		pos = float4(front, 0);
    }
    else
    {
		direction = normalize(front - back);
		pos = float4(back, 0);
	}
 
    //compositing step      
    float4 oldValue = 0;
    float4 newValue = 0;
    float4 finalValue = 0;
    
    float value = 0.0f;
	
	float3 Step = direction * StepSize;
	
	float adaptiveStep = 1.0f;
    
    for(int i = 0; i < Iterations; i++)
    {
		pos.w = 0;
		// get density value
		value = tex3Dlod(VolumeSampler, pos).a;
		
		if(shading)
		{
		    //get gradient
			float3 gradient = tex3Dlod(VolumeSampler, pos).rgb;
			gradient.xyz = normalize(gradient.xyz*2.0f - 1.0f);
			//float magnitude = length(gradient);
			//gradient = float3(magnitude, magnitude, magnitude);

			//float3 LightDirection = normalize(DirectLightVector);
		    float3 ViewDirection = normalize(input.viewDirection);
			float3 LightDirection = normalize(DirectLightVector - pos.xyz); 
			//float3 ViewDirection = normalize(CameraPosition - pos.xyz);
			float3 Half = normalize(LightDirection + ViewDirection);
			float diff = saturate(dot(LightDirection, gradient));
					
				
			if(useTransferColors)
			{
			   //get color from cvd transferTexture
			   if(useCVDTransferColors)
			   {
			     newValue = tex1Dlod(CVDTransferSampler, value);
			   }
			   else
			   { 
				 //get color from transferTexture
				 newValue = tex1Dlod(TransferSampler, value);
				}
			}
			else
			{
				// density values produce gray picture
				newValue = float4(value, value, value, value);
			}

			//calculate color for ambient lighting
			float3 AmbientLight = CalculateAmbient(newValue, AmbientLightColor);
			
			//calculate diffuse lighting
			float3 DiffuseLight = CalculateDiffuse(newValue, DirectLightColor, diff);
			
			//calculate specular Lighting 
			//phong
			//float3 Reflect = normalize(2 * diff * gradient - LightDirection);
			//float3 SpecularLight = SpecularLightColor * pow(saturate(dot(Reflect, ViewDirection)), 64.0f);
				
		    //blinn
			float specularTerm =  pow(saturate(dot(Half, gradient)), shininess);
			float3 SpecularLight = SpecularLightColor * specularTerm;

			//combine the light-values
			newValue.rgb = LightingCombine(AmbientLight, DiffuseLight, SpecularLight);
			 // newValue.rgb = LightingCombine(AmbientLight, float3(0.0f, 0.0f, 0.0f), float3(0.0f, 0.0f, 0.0f));
				
		}
		else
		{			
			if(useTransferColors)
			{
			   //get color from cvd transferTexture
			   if(useCVDTransferColors)
			   {
			     newValue = tex1Dlod(CVDTransferSampler, value);
			   }
			   else
			   { 
				 // get color from transferTexture
				 newValue = tex1Dlod(TransferSampler, value);
			   }
			}
			else
			{
				// density values produce gray picture
				newValue = float4(value, value, value, value);
			}
		}
		
		if(useAlpha)
		{
			// alpha from transferfunction
			newValue.a = tex1Dlod(TransferSampler, value).a;

			//make it more transparent	
			//newValue.a *= 0.5f;
			
			//Oppacity correction: As sampling distance decreases we get more samples.
			//Therefore the alpha values set for a sampling distance of .5f will be too
			//high for a sampling distance of .25f (or conversely, too low for a sampling
			//distance of 1.0f). So we have to adjust the alpha accordingly.
			//newValue.a = 1 - pow((1 - newValue.a), StepSize / 0.5f);
		}
		else
		{
			// set alpha from density values
			newValue.a = value;
			
			//make it more transparent	TODO: alpha from transferfunction	
			//newValue.a *= 0.1f; //src.a *= 0.02f; //default = 1.0f;
						  //reduce the alpha to have a more transparent result
						  //TODO: this needs to be adjusted based on the step size
						  //i.e. the more steps we take, the faster the alpha will grow	
		}

		//Oppacity correction: As sampling distance decreases we get more samples.
		//Therefore the alpha values set for a sampling distance of .5f will be too
		//high for a sampling distance of .25f (or conversely, too low for a sampling
		//distance of 1.0f). So we have to adjust the alpha accordingly.
		newValue.a = 1 - pow((1 - newValue.a), sampleDist / baseSampleDist); 


		//color composition 		
		if(frontToBack)
		{					
			//front to back blending
			finalValue.rgb = newValue.rgb * newValue.a *(1 - oldValue.a) + oldValue.rgb;
			finalValue.a = newValue.a *(1 - oldValue.a) + oldValue.a; 

		}
		else
		{
			//back to front blending ... ATTENTION: you may not stop at alpha > 0.99 for back to front!
			finalValue.rgb = newValue.rgb * newValue.a + oldValue.rgb * (1 - newValue.a);			
			finalValue.a = newValue.a + oldValue.a *(1 - newValue.a); 

	     }
		
		oldValue = finalValue;

		//set iteration
		if(value > 0.0f)
		{
			adaptiveStep = 1.0f;
			//adaptiveStep = newValue.a*5.0f;
			//adaptiveStep = 1.0f/(value*10.0f);
		}
		else
		{
			adaptiveStep = 1.0f;
		}
		
		//break from the loop when alpha gets high enough
		if((finalValue.a >= 0.95f) && (frontToBack == true))
		{
			i = Iterations;	
		}
		else
			{
			
			//advance the current position
			pos.xyz += Step * adaptiveStep;
			
			//break if the position is greater than <1, 1, 1>
			if(pos.x > 1.0f || pos.y > 1.0f || pos.z > 1.0f)
			{
				i = Iterations;
			}
			
		}
    }    
    return finalValue;
}

float4 MaxIntensityProjPS(VertexShaderOutput input) : COLOR0
{ 
	//calculate projective texture coordinates
	//used to project the front and back position textures onto the cube
	float2 texC = input.pos.xyz /= input.pos.w;
	texC.x =  0.5f*texC.x + 0.5f; 
	texC.y = -0.5f*texC.y + 0.5f;  
	
    float3 front = tex2D(FrontSampler, texC).xyz;
    float3 back = tex2D(BackSampler, texC).xyz;    
   
	float3 direction = (back - front);
	direction = normalize(direction);
	float4 pos = float4(front, 0);
            
    float4 finalValue = 0;    
    float4 newValue = 0;
    
    float value = 0.0f;
	
	float3 Step = direction * StepSize;
	
	float3 gradient = 0.0f;
    
    for(int i = 0; i < Iterations; i++)
    {
		pos.w = 0;
		
		// get density value 
		value = tex3Dlod(VolumeSampler, pos).a;
		
		if(useAlpha)
		{
			// alpha from transferfunction
			newValue.a = tex1Dlod(TransferSampler, value).a;
			//make it more transparent	
			newValue.a *= 0.5f;			
		}
		else
		{
			// set alpha from density values
			newValue.a = value;
			
		}
		
		if(newValue.a > finalValue.a)
		{	
			if(useTransferColors)
			{
				// get color from transferTexture
				finalValue = tex1Dlod(TransferSampler, newValue.a);
			}
			else
			{
				// density values produce gray picture
				finalValue = float4(newValue.a, newValue.a, newValue.a, newValue.a);
			}
			
			finalValue.rgb *= newValue.a;
			
			// set alpha
			finalValue.a = newValue.a;
			
			gradient = tex3Dlod(VolumeSampler, pos).rgb;
		
		}
		
		//advance the current position
		pos.xyz += Step;
			
		//break if the position is greater than <1, 1, 1>
		if(pos.x > 1.0f || pos.y > 1.0f || pos.z > 1.0f)
		{
			i = Iterations;
		}
    }  
    
    if(shading)
		{
			
			float3 LightDirection = normalize(DirectLightVector); 
			float3 ViewDirection = normalize(input.viewDirection);
			float3 diff = saturate(dot(gradient, LightDirection));			
				
			if(useTransferColors)
			{
				// get color from transferTexture
				newValue = tex1Dlod(TransferSampler, finalValue.a);
			}
			else
			{
				// density values produce gray picture
				newValue = float4(finalValue.a, finalValue.a, finalValue.a, finalValue.a);
			}
		
			//calculate color for ambient lighting
			float3 AmbientLight = CalculateAmbient(newValue, AmbientLightColor);
			AmbientLight *= finalValue.a;
			
			//calculate diffuse lighting
			float3 DiffuseLight = CalculateDiffuse(newValue, DirectLightColor, diff);
			//DiffuseLight *= finalValue.a;
			
			//calculate specular Lighting 
			float3 Reflect = normalize(2 * diff * gradient - LightDirection);
			float SpecularLight = pow(saturate(dot(Reflect, ViewDirection)), 64.0f);
			SpecularLight *= finalValue.a;
			
			//combine the light-values
			newValue.rgb = LightingCombine(AmbientLight, DiffuseLight, SpecularLight);
			finalValue = newValue;
		}   
      
    return finalValue;
}

//----------------------------------------------------------------------------------
//Render Scene
//----------------------------------------------------------------------------------
technique RenderPosition
{
    pass Pass1
    {		
        VertexShader = compile vs_2_0 PositionVS();
        PixelShader = compile ps_2_0 PositionPS();
    }
}

technique RenderSlice
{
    pass Pass1
    {		
        VertexShader = compile vs_3_0 PositionVS();
        PixelShader = compile ps_3_0 SlicePS();
    }
}

technique RayCastDirection
{
    pass Pass1
    {		
        VertexShader = compile vs_2_0 PositionVS();
        PixelShader = compile ps_2_0 DirectionPS();
    }
}

technique RayCastSimple
{
    pass Pass1
    {		
        VertexShader = compile vs_3_0 PositionVS();
        PixelShader = compile ps_3_0 RayCastSimplePS();
    }
}

technique MaximumIntensity
{
    pass Pass1
    {		
        VertexShader = compile vs_3_0 PositionVS();
        PixelShader = compile ps_3_0 MaxIntensityProjPS();
    }
}

technique WireFrame
{
    pass Pass1
    {		
        VertexShader = compile vs_2_0 PositionVS();
        PixelShader = compile ps_2_0 WireFramePS();
    }
}
