﻿#region Using Statements
using System;
using System.Threading;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
#endregion

#pragma warning disable 67

namespace visLU
{
    /// <summary>
    /// Class for creating and managing the GraphicsDevice, used by the windows forms controls
    /// implements the IGraphicsDeviceService, that is responsible for retrieving GraphicsDevice objects.
    /// </summary>
    class VisGraphicsDevice: IGraphicsDeviceService
    {

        #region Variables

        //GraphicsDevice instance
        static VisGraphicsDevice instance;
        //how many controls are sharing the graphics device
        static int referenceCount;

        GraphicsDevice graphicsDevice;
        // Store the current device settings.
        PresentationParameters parameters;
        //GraphicsDeviceManager graphics;//todo: do we need this? + setGraphicsDeviceManager(); ->(Game1.cs)

        //IGraphicsDeviceService events
        public event EventHandler<EventArgs> DeviceCreated;
        public event EventHandler<EventArgs> DeviceDisposing;
        public event EventHandler<EventArgs> DeviceReset;
        public event EventHandler<EventArgs> DeviceResetting;

        #endregion

        #region Properties
        /// <summary>
        /// Gets the current graphics device.
        /// </summary>
        public GraphicsDevice GraphicsDevice
        {
            get { return graphicsDevice; }
        }

       
        #endregion
       
        #region Constructor

        VisGraphicsDevice(IntPtr windowHandle, int width, int height)
        {
            parameters = new PresentationParameters();

            parameters.BackBufferWidth = Math.Max(width, 1);
            parameters.BackBufferHeight = Math.Max(height, 1);
            parameters.BackBufferFormat = SurfaceFormat.Rgba64;
            parameters.DepthStencilFormat = DepthFormat.Depth24;
            parameters.DeviceWindowHandle = windowHandle;
            //parameters.MultiSampleCount = 8;
            parameters.PresentationInterval = PresentInterval.Default; 
            parameters.IsFullScreen = false;
            parameters.MultiSampleCount = 0;

            graphicsDevice = new GraphicsDevice(GraphicsAdapter.DefaultAdapter,
                                                GraphicsProfile.HiDef,
                                                parameters);
        }

        #endregion

        /// <summary>
        /// Adds a reference to the VisGraphicsDevice instance.
        /// Creates the instance, if called for first time. 
        /// </summary>
        public static VisGraphicsDevice Add(IntPtr windowHandle, int width, int height)
        {

            if (Interlocked.Increment(ref referenceCount) == 1)
            {
                instance = new VisGraphicsDevice(windowHandle, width, height);
            }

            return instance;
        }


        /// <summary>
        /// Releases a reference to the VisGraphicsDevice instance.
        /// </summary>
        public void Release(bool disposing)
        {
            //Decrement the control references count to the VisGraphicsDevice instance
            if (Interlocked.Decrement(ref referenceCount) == 0)
            {
                // Dispose the VisGraphicsDevice instance if no more control references
                if (disposing)
                {
                    if (DeviceDisposing != null)
                        DeviceDisposing(this, EventArgs.Empty);

                    graphicsDevice.Dispose();
                }

                graphicsDevice = null;
            }
        }


        /// <summary>
        /// Resets the graphics device to whichever is bigger out of the specified
        /// resolution or its current size. This behavior means the device will
        /// demand-grow to the largest of all its GraphicsDeviceControl clients.
        /// </summary>
        public void ResetDevice(int width, int height)
        {
            if (DeviceResetting != null)
                DeviceResetting(this, EventArgs.Empty);

            parameters.BackBufferWidth = Math.Max(parameters.BackBufferWidth, width);
            parameters.BackBufferHeight = Math.Max(parameters.BackBufferHeight, height);

            graphicsDevice.Reset(parameters);

            if (DeviceReset != null)
                DeviceReset(this, EventArgs.Empty);
        }


    }
 
}